from typing import Any
import datetime


from aw_client.models.model_schema import ModelObject, ModelObjectField
from .test_data import ModelObjectTestData



try:
    from pyspark.sql import SparkSession, DataFrame
    from pyspark.sql.types import DataType, StringType, DoubleType, TimestampType, LongType, BooleanType, \
        ByteType, ShortType, IntegerType, DecimalType, FloatType, DateType, StructType, StructField
except ImportError:
    raise Exception('Для использования Spark установите библиотеку с опцией [dev]: `pip install analytic-workspace-client[dev]`')


def build_spark_schema(test_data: ModelObjectTestData) -> StructType:
    """ """
    if test_data.schema is not None:
        return StructType(fields=[
            StructField(sf['model_name'], spark_type_for_simple_type(sf['simple_type']), True) for sf in test_data.schema
        ])
    return StructType(fields=[
        StructField(n, spark_type_for_python_value(v), True) for n, v in test_data.rows[0].items()
    ])


def build_dataframe(spark: SparkSession, test_data: ModelObjectTestData) -> DataFrame:
    """ """
    return spark.createDataFrame(test_data.rows, schema=build_spark_schema(test_data))


def build_model_object(test_data: ModelObjectTestData):
    """ """
    if test_data.schema is not None:
        fields = [
            ModelObjectField(name=f['model_name'], model_name=f['model_name'], simple_type=f['simple_type']) for f in test_data.schema
        ]
    else:
        fields = [
            ModelObjectField(name=n, model_name=n, simple_type=simple_type_for_python_value(v)) for n, v in test_data.rows[0].items()
        ]
    return ModelObject(
        name=test_data.model_name,
        model_name=test_data.model_name,
        type='table',
        sql_text=None,
        fields=fields
    )




def spark_type_for_simple_type(simple_type: str) -> DataType:
    """ 
    """
    if simple_type == 'number':
        return LongType()
    elif simple_type == 'float':
        return DoubleType()
    elif simple_type == 'date':
        return TimestampType()
    elif simple_type == 'bool':
        return BooleanType()
    else:
        return StringType()
        

def spark_type_for_python_value(value: Any) -> DataType:
    """ 
    Returns Spark type
    """
    if isinstance(value, datetime.date):
        return DateType()
    if isinstance(value, datetime.datetime):
        return TimestampType()
    if isinstance(value, int):
        return LongType()
    if isinstance(value, float):
        return DoubleType()
    if isinstance(value, bool):
        return BooleanType()
    return StringType()


def simple_type_for_python_value(value: Any) -> str:
    """ """
    if isinstance(value, datetime.date):
        return 'date'
    if isinstance(value, datetime.datetime):
        return 'date'
    if isinstance(value, int):
        return 'number'
    if isinstance(value, float):
        return 'float'
    if isinstance(value, bool):
        return 'bool'
    return 'string'


import shutil
from typing import Optional

from pathlib import Path

import pandas

from aw_client.data_master import DataMasterApi
from aw_client.domain import APIConfig
from aw_client.models.model_schema import ModelSchema

from .v1 import DataMasterV1
from ..exceptions import AwModelForbidden
from ..tools import get_temp_folder


class DataMasterV2(DataMasterApi):
    """ """
    def __init__(self, api_config: APIConfig):
        super().__init__(api_config)

        self.api_config = api_config
        self.api_v1 = DataMasterV1(api_config)  # это не очень хорошо, но пока идей нет, как лучше сделать

    def load_model(self, model_id: int, **options) -> pandas.DataFrame:
        return self.api_v1.load_model(model_id, **options)

    def model_schema(self, model_id: int) -> ModelSchema:
        """ """
        with self.get_http_client() as client:
            r = client.get(url='data-master/v2/model/schema', params={'model_id': model_id})
            if not r.is_success:
                if r.status_code == 403:
                    raise AwModelForbidden(f'Доступ к модели {model_id} запрещен')
                raise Exception(f'Ошибка запроса схемы модели HTTP {r.status_code} GET {r.url}: {r.text}')

            return ModelSchema.parse_obj(r.json())

    def load_model_object_data(self, folder: Path, model_id: int, model_object_name: str, limit: Optional[int]):
        """ """
        with self.get_http_client() as client:
            params = {
                'model_id': model_id,
                'model_object_name': model_object_name
            }
            if limit is not None:
                params['limit'] = limit

            with get_temp_folder() as temp_folder:
                temp_file = temp_folder / f'{model_object_name}.zip'
                with client.stream(method='GET', url='data-master/v2/model-object/data', params=params, timeout=None) as r:
                    if not r.is_success:
                        r.read()
                        raise DataMasterApi.Error(f'Ошибка запроса GET {r.url}: HTTP {r.status_code}: {r.text}')
                    else:
                        with open(temp_file, 'wb') as f:
                            for chunk in r.iter_bytes():
                                f.write(chunk)

                if (folder / f'{model_object_name}.parquet').exists():
                    shutil.rmtree(folder / f'{model_object_name}.parquet')

                shutil.unpack_archive(temp_file, folder.as_posix())

        return



from typing import Dict, Optional, Any, Union, List, Literal
from pathlib import Path
from collections import OrderedDict
import datetime
import inspect

from aw_client.core.compiler import ScriptCompiler
from aw_client.core.model_vault import Vault
from aw_client.core.spark import build_spark_session
from aw_client.core.bundle import NamedObjectsBundle
from aw_client.models.model_schema import ModelObject, ModelObjectField
from .application import ETLBlockApplication
from .test_data import ModelObjectTestData
from .tools import build_dataframe, build_model_object, build_spark_schema
from .dto import ETLBlockMeta


try:
    from pyspark.sql import SparkSession, DataFrame
    from pyspark.sql.types import DataType, StringType, DoubleType, TimestampType, LongType, BooleanType, \
        ByteType, ShortType, IntegerType, DecimalType, FloatType, DateType, StructType, StructField
except ImportError:
    raise Exception('Для использования Spark установите библиотеку с опцией [dev]: `pip install analytic-workspace-client[dev]`')



def get_etl_block_meta(block_path: Path) -> ETLBlockMeta:
    """ """
    block_meta_path = block_path / 'block_meta.json' if block_path.is_dir() else block_path

    if not block_meta_path.exists():
        raise Exception(f'Файл с метаданным блока не найден: {block_meta_path}')

    with open(block_meta_path, 'rt') as f:
        return ETLBlockMeta.parse_raw(f.read())


def get_etl_block_schema(block_path: Path, 
                         test_data: Union[ModelObjectTestData, List[ModelObjectTestData]],
                         params: Optional[Dict] = None, 
                         run_mode: Optional[Literal['']] = None, 
                         vault: Optional[Vault] = None,
                         model_script_code: Optional[str] = None) -> StructType:
    """ 
    Args:

    """
    block_code_path = block_path / 'block_code.py' if block_path.is_dir() else block_path
    
    if not block_code_path.exists():
        raise Exception(f'Файл с исходным кодом блока не найден: {block_code_path}')
    
    with open(block_code_path, 'rt') as f:
        block_code = f.read()

    # Компиляция кода блока
    try:
        block_module = ScriptCompiler().compile(source_code=block_code, mode=ScriptCompiler.MODE_ETL_BLOCK)
    except ScriptCompiler.CannotCompile as e:
        raise Exception(f'Ошибка компиляции исходного кода блока: {e}')
    
    # Компиляция кода модели
    if model_script_code:
        try:
            model_module = ScriptCompiler().compile(source_code=model_script_code, mode=ScriptCompiler.MODE_ETL)
        except ScriptCompiler.CannotCompile as e:
            raise Exception(f'Ошибка компиляции исходного кода скрипта модели: {e}')
    else:
        model_module = None

    spark = build_spark_session()

    # Дочерние датафреймы
    dataframes = OrderedDict()
    for td in (test_data if isinstance(test_data, list) else [test_data]):
        df = build_dataframe(spark, td)
        if not dataframes:
            dataframes['child'] = df
    
    upstream_dataframes = NamedObjectsBundle(dataframes)

    # Дочерние схемы
    schemas = OrderedDict()
    for td in (test_data if isinstance(test_data, list) else [test_data]):
        schema = build_spark_schema(td)
        if not schemas:
            schemas['child'] = schema
        schemas[td.model_name] = schema

    upstream_schemas = NamedObjectsBundle(schemas)
    
    block_schema_parameters = inspect.signature(block_module['block_schema']).parameters

    app = ETLBlockApplication(
        spark_builder=build_spark_session,
        run_mode=run_mode or 'full',
        vault=vault or Vault(),
        model_module=model_module
    )

    # Определение параметров для передачи
    block_schema_kwargs = {}
    if 'params' in block_schema_parameters:
        block_schema_kwargs['params'] = params
    if 'app' in block_schema_parameters:
        block_schema_kwargs['app'] = app
    if 'model_object' in block_schema_parameters:
        block_schema_kwargs['model_object'] = build_model_object(test_data)
    if 'schema' in block_schema_parameters:
        block_schema_kwargs['schema'] = upstream_schemas.first()
    if 'schemas' in block_schema_parameters:
        block_schema_kwargs['schemas'] = upstream_schemas
    if 'upstream_schema' in block_schema_parameters:
        block_schema_kwargs['upstream_schema'] = upstream_schemas.first()
    if 'upstream_schemas' in block_schema_parameters:
        block_schema_kwargs['upstream_schemas'] = upstream_schemas
    if 'df' in block_schema_parameters:
        block_schema_kwargs['df'] = upstream_dataframes.first()
    if 'dfs' in block_schema_parameters:
        block_schema_kwargs['dfs'] = upstream_dataframes
    if 'upstream_dataframe' in block_schema_parameters:
        block_schema_kwargs['upstream_dataframe'] = upstream_dataframes.first()
    if 'upstream_dataframes' in block_schema_parameters:
        block_schema_kwargs['upstream_dataframes'] = upstream_dataframes
    

    return block_module['block_schema'](**block_schema_kwargs)
    

def get_etl_block_data(block_path: Path, 
                       test_data: Union[ModelObjectTestData, List[ModelObjectTestData]],
                       params: Optional[Dict] = None, 
                       run_mode: Optional[Literal['']] = None, 
                       vault: Optional[Vault] = None,
                       model_script_code: Optional[str] = None) -> StructType:
    """ 
    Args:

    """
    block_code_path = block_path / 'block_code.py' if block_path.is_dir() else block_path
    
    if not block_code_path.exists():
        raise Exception(f'Файл с исходным кодом блока не найден: {block_code_path}')
    
    with open(block_code_path, 'rt') as f:
        block_code = f.read()

    # Компиляция кода блока
    try:
        block_module = ScriptCompiler().compile(source_code=block_code, mode=ScriptCompiler.MODE_ETL_BLOCK)
    except ScriptCompiler.CannotCompile as e:
        raise Exception(f'Ошибка компиляции исходного кода блока: {e}')
    
    # Компиляция кода модели
    if model_script_code:
        try:
            model_module = ScriptCompiler().compile(source_code=model_script_code, mode=ScriptCompiler.MODE_ETL)
        except ScriptCompiler.CannotCompile as e:
            raise Exception(f'Ошибка компиляции исходного кода скрипта модели: {e}')
    else:
        model_module = None

    spark = build_spark_session()

    # Дочерние датафреймы
    dataframes = OrderedDict()
    for td in (test_data if isinstance(test_data, list) else [test_data]):
        df = build_dataframe(spark, td)
        if not dataframes:
            dataframes['child'] = df
        dataframes[td.model_name] = df
    
    upstream_dataframes = NamedObjectsBundle(dataframes)

    block_data_parameters = inspect.signature(block_module['block_data']).parameters

    app = ETLBlockApplication(
        spark_builder=build_spark_session,
        run_mode=run_mode or 'full',
        vault=vault or Vault(),
        model_module=model_module
    )

    # Определение параметров для передачи
    block_schema_kwargs = {}
    if 'params' in block_data_parameters:
        block_schema_kwargs['params'] = params
    if 'app' in block_data_parameters:
        block_schema_kwargs['app'] = app
    if 'model_object' in block_data_parameters:
        block_schema_kwargs['model_object'] = build_model_object(test_data)
    if 'df' in block_data_parameters:
        block_schema_kwargs['df'] = upstream_dataframes.first()
    if 'dfs' in block_data_parameters:
        block_schema_kwargs['dfs'] = upstream_dataframes
    if 'upstream_dataframe' in block_data_parameters:
        block_schema_kwargs['upstream_dataframe'] = upstream_dataframes.first()
    if 'upstream_dataframes' in block_data_parameters:
        block_schema_kwargs['upstream_dataframes'] = upstream_dataframes
    
    return block_module['block_data'](**block_schema_kwargs)
    



from typing import Optional

import os
from pathlib import Path

import httpx
import pandas
from contextlib import contextmanager
from urllib.parse import urljoin

from aw_client.data_master import get_nearest_api
from aw_client.domain import APIConfig
from aw_client.models.model_schema import ModelSchema
from aw_client.exceptions import AwClientMisconfigured

from aw_client.tools import strip_column_name_prefix


class Session:
    """
    Сессия подключения к AW
    """
    class Error(Exception):
        """ """

    def __init__(self, token: str = None, aw_url: str = None, version: int = None):
        """ """
        self.token = (token or os.getenv("AW_DATA_TOKEN") or '').strip()
        self.aw_url = (aw_url or os.getenv('AW_URL') or '').strip()
        if not self.aw_url:
            raise AwClientMisconfigured(
                'Укажите URL к AnalyticWorkspace: Session(aw_url=\'http://aw.mydomain.ru\') или установите переменную '
                'окружения AW_URL')
        if not self.aw_url.endswith('/'):
            self.aw_url += '/'
        self.version = version if version is not None else None

    def get_aw_version(self) -> int:
        """ """
        if self.version is None:
            with self.get_aw_client() as client:
                r = client.get('data-master/_version')
                if r.is_success:
                    self.version = r.json().get('version')
                elif r.status_code == 404:
                    self.version = 0  # для версии 0 как раз нет
                else:
                    raise Session.Error(f'Ошибка HTTP {r.status_code} запроса GET {r.url}: {r.text}')
        if self.version is None:
            raise Session.Error('Не удалось определить версию AW')

        return self.version

    def load_model(self, model_id: int, strip_prefix: bool = False, **options) -> pandas.DataFrame:
        """
        Загружает данные модели в датафрейм pandas
        """
        if not self.aw_url:
            raise Session.Error('Не указан URL к Analytic Workspace (параметр aw_url)')

        if not self.token:
            raise Session.Error('Не указан токен доступа к Analytic Workspace (параметр token)')

        api_config = APIConfig(
            aw_url=self.aw_url,
            token=self.token
        )

        api, version = get_nearest_api(aw_version=self.get_aw_version(), api_config=api_config)
        if api is None:
            raise Session.Error(f'Не удалось получить API для версии AW {self.version}')

        df = api.load_model(model_id=model_id, **options)
        if strip_prefix:
            df = strip_column_name_prefix(df, inplace=True)

        return df

    def model_schema(self, model_id: int) -> ModelSchema:
        """ """
        if not self.aw_url:
            raise Session.Error('Не указан URL к Analytic Workspace (параметр aw_url)')

        if not self.token:
            raise Session.Error('Не указан токен доступа к Analytic Workspace (параметр token)')

        api_config = APIConfig(
            aw_url=self.aw_url,
            token=self.token
        )

        api, version = get_nearest_api(aw_version=self.get_aw_version(), api_config=api_config)
        if api is None:
            raise Session.Error(f'Не удалось получить API для версии AW {self.version}')

        return api.model_schema(model_id)

    def load_model_object_data(self, folder: Path, model_id: int, model_object_name: str, limit: Optional[int]):
        """ """
        if not self.aw_url:
            raise Session.Error('Не указан URL к Analytic Workspace (параметр aw_url)')

        if not self.token:
            raise Session.Error('Не указан токен доступа к Analytic Workspace (параметр token)')

        api_config = APIConfig(
            aw_url=self.aw_url,
            token=self.token
        )

        api, version = get_nearest_api(aw_version=self.get_aw_version(), api_config=api_config)
        if api is None:
            raise Session.Error(f'Не удалось получить API для версии AW {self.version}')

        return api.load_model_object_data(folder, model_id, model_object_name, limit)

    @contextmanager
    def get_aw_client(self) -> httpx.Client:
        """ """
        if not self.aw_url:
            raise Session.Error('Не указан URL к Analytic Workspace')

        headers = {"Authorization": f"Bearer {self.token}"}

        client = httpx.Client(base_url=self.aw_url, headers=headers)

        try:
            yield client
        finally:
            client.close()

    @property
    def mlflow(self):
        """ """
        try:
            import mlflow
        except ImportError:
            raise AwClientMisconfigured(
                'Для использованиея MLFlow установите библиотеку с опцией ml: pip install analytic-workspace-client[ml]')

        if not self.token:
            data_master_url = urljoin(self.aw_url, 'data-master/get-token')
            raise AwClientMisconfigured(
                f'Не указан токен доступа к AnalyticWorkspace. Пройдите по адресу {data_master_url} для получения '
                f'токена')

        tracking_url = urljoin(self.aw_url, 'mlflow')

        if mlflow.get_tracking_uri() != tracking_url:
            mlflow.set_tracking_uri(tracking_url)

        os.environ['MLFLOW_TRACKING_TOKEN'] = self.token

        return mlflow



from typing import Optional, Callable

try:
    from pyspark.sql import SparkSession
except ImportError:
    raise Exception('Для использования Spark установите библиотеку с опцией [dev]: `pip install analytic-workspace-client[dev]`')

from aw_client.core.model_vault import Vault
from aw_client.core.compiler import CompiledModule


class ETLBlockApplication:
    """ """
    def __init__(self, 
                 spark_builder: Callable, 
                 run_mode: str, 
                 vault: Vault,
                 model_module: Optional[CompiledModule] = None):
        self._spark_builder = spark_builder
        self._spark = None
        self._run_model = run_mode
        self._model_module = model_module
        self._vault = vault
    
    @property
    def spark(self) -> SparkSession:
        if self._spark is None:
            self._spark = self._spark_builder()
        return self._spark
    
    @property
    def is_spark_initialized(self) -> bool:
        return self._spark is not None
    
    @property
    def model_module(self) -> Optional[CompiledModule]:
        """ """
        return self._model_module

    @property
    def vault(self) -> Vault:
        """ """
        return self._vault

    @property
    def run_mode(self) -> str:
        """ """
        return self.run_mode
    


from typing import Tuple, Optional, Callable

from importlib import import_module

from .base import DataMasterApi
from ..domain import APIConfig


def get_nearest_api(aw_version: int, api_config: APIConfig, exact_version: bool = False) -> Tuple[Optional[DataMasterApi], Optional[int]]:
    """ """
    api_cls = None
    version = None

    # передираем все доступные APi
    for version in range(aw_version, (aw_version - 1) if exact_version else -1, -1):
        module_name = f'aw_client.data_master.v{version}'
        class_name = f'DataMasterV{version}'

        try:
            api_mdl = import_module(module_name)
            api_cls = getattr(api_mdl, class_name, None)
        except ImportError:
            continue

        if api_cls is not None:
            break

    return api_cls(api_config) if api_cls is not None else None, version



from typing import Optional, Dict, Any


class Vault:
    """ """
    def __init__(self, values: Optional[dict] = None):
        self._values = values or {}

    @property
    def values(self) -> dict:
        return self._values

    def get(self, name: str) -> Optional[Any]:
        return self._values.get(name, None)



from typing import Protocol


class CompiledModule(Protocol):
    """ """
    def __contains__(self, item):
        """ """

    def __getattr__(self, item):
        """ """

    def __getitem__(self, item):
        """ """

    @property
    def is_empty(self) -> bool:
        """ """


import tempfile
import shutil
from pathlib import Path
from contextlib import contextmanager

import pandas as pd


@contextmanager
def get_temp_folder(remove_on_exit: bool = True) -> Path:
    """ """
    temp_folder = tempfile.mkdtemp()

    try:
        yield Path(temp_folder)
    finally:
        if remove_on_exit:
            shutil.rmtree(temp_folder, ignore_errors=True)


def strip_column_name_prefix(df: pd.DataFrame, inplace: bool = False) -> pd.DataFrame:
    """ 
    Вырезает общий префикс у столбцов датафрейма 
    """
    column_renames = {}
    for column_name in df.columns:
        if '__' in column_name:
            _, column_renames[column_name] = column_name.rsplit('__', 1)

    if column_renames:
        if inplace:
            df.rename(columns=column_renames, inplace=inplace)
        else:
            df.rename(columns=column_renames)

    return df



from hashlib import md5
from urllib.parse import urlparse
from pathlib import Path

from aw_client.models.model_schema import ModelSchema, ModelObject


class ModelDevCache:
    """ """
    def __init__(self, cache_root_folder: Path, aw_url: str, model_id: int, is_full: bool = False):
        """ """
        self.cache_root_folder = cache_root_folder
        self.aw_url = aw_url
        self.model_id = model_id
        self.is_full = is_full

        self.aw_netloc = urlparse(aw_url).netloc.replace(":", "_")

    def get_model_folder(self) -> Path:
        """ """
        return Path(self.cache_root_folder) / self.aw_netloc / f'model_{self.model_id}'

    def model_schema_path(self) -> Path:
        return self.get_model_folder() / 'model_schema.json'

    def model_source_folder(self) -> Path:
        """ """
        return self.get_model_folder() / ('full' if self.is_full else 'part')

    def model_object_data_folder(self, model_object_name: str) -> Path:
        """ """
        return self.model_source_folder() / f'{model_object_name}.parquet'

    def model_object_hash_path(self, model_object: ModelObject) -> Path:
        """ """
        return self.model_source_folder() / f'{model_object.model_name}.hash'

    def set_model_object_hash(self, model_object: ModelObject):
        """ """
        with open(self.model_object_hash_path(model_object), 'wt') as f:
            f.write(self._model_object_hash(model_object))

    def is_model_object_actual(self, model_object: ModelObject) -> bool:
        """ """
        try:
            model_object_hash_path = self.model_object_hash_path(model_object)
            if not model_object_hash_path.exists():
                return False

            with open(self.model_object_hash_path(model_object), 'rt') as f:
                return f.read() == self._model_object_hash(model_object)
        except:
            return False

    @staticmethod
    def _model_object_hash(model_object: ModelObject) -> str:
        return md5(model_object.json(ensure_ascii=False, sort_keys=True).encode()).hexdigest()




try:
    from pyspark.sql import SparkSession
except ImportError:
    raise Exception('Для использования Spark установите библиотеку с опцией [dev]: `pip install analytic-workspace-client[dev]`')


def build_spark_session():
    """ """
    return SparkSession.builder \
        .master('local[*]') \
        .config('spark.driver.host', '127.0.0.1') \
        .config('spark.ui.enabled', 'false') \
        .getOrCreate()




class DataMasterApiError(Exception):
    """
    Исключения, которые выдаются API датамастера
    """


class WrongDataMasterApiVersion(DataMasterApiError):
    """
    Попытка использовать API датамастера, которое несовместимо с API AW
    """


class AwModelForbidden(DataMasterApiError):
    """
    Доступ к модели запрещен
    """


class AwClientMisconfigured(Exception):
    """ """



from typing import Optional
from pathlib import Path

import pandas

from aw_client.data_master.base import DataMasterApi
from aw_client.exceptions import WrongDataMasterApiVersion
from aw_client.models.model_schema import ModelSchema
from aw_client.tools import get_temp_folder


class DataMasterV1(DataMasterApi):
    """ """
    def load_model(self, model_id: int, **options) -> pandas.DataFrame:
        """ """
        apply_schema = options.get('schema')
        if apply_schema is None:
            apply_schema = True

        timeout = options.get('timeout')
        if timeout is not None:
            timeout = int(timeout)

        with self.get_http_client() as client:

            model_schema = None
            if apply_schema:
                model_schema = client.get('data-master/v1/model/schema', params={'model_id': model_id}).json()

            with get_temp_folder() as temp_folder:
                temp_file = temp_folder / 'model_data.json'
                params = {'model_id': model_id, 'format': 'json-lines'}
                with client.stream(method='GET', url='data-master/v1/model/data', params=params, timeout=timeout) as r:
                    if not r.is_success:
                        r.read()
                        raise DataMasterApi.Error(f'Ошибка запроса GET {r.url}: HTTP {r.status_code}: {r.text}')
                    with open(temp_file, 'wb') as f:
                        for chunk in r.iter_bytes():
                            f.write(chunk)

                    df = pandas.read_json(temp_file, lines=True)

        if apply_schema:
            as_type = {}

            for column in model_schema['fields']:
                # if column['simple_type'] == 'string':
                #     as_type[column['model_name']] = 'object'
                # elif column['simple_type'] == 'number':
                #     as_type[column['model_name']] = 'int64'
                # elif column['simple_type'] == 'float':
                #     as_type[column['model_name']] = 'float64'
                # elif column['simple_type'] == 'bool':
                #     as_type[column['model_name']] = 'bool'
                if column['simple_type'] == 'date':
                    as_type[column['model_name']] = 'datetime64[ns]'

            df = df.astype(as_type)

        return df

    def model_schema(self, model_id: int) -> ModelSchema:
        raise WrongDataMasterApiVersion(f'Получение схемы модели доступно для Analytic Workspace с версии 1.19')

    def load_model_object_data(self, folder: Path, model_id: int, model_object_name: str, limit: Optional[int]):
        """ """
        raise WrongDataMasterApiVersion(f'Получение данных объекта модели доступно для AW с версии 1.19')



from typing import Optional, List, Dict

from pydantic import BaseModel


class ModelObjectField(BaseModel):
    """ """
    name: str
    model_name: str
    simple_type: str


class ModelObject(BaseModel):
    """ """
    name: str
    model_name: str
    type: str
    sql_text: Optional[str] = None
    fields: List[ModelObjectField] = []


class ModelSql(BaseModel):
    """
    """
    sql: str
    xsqls: Dict[str, str] = {}


class ModelSchema(BaseModel):
    """ """
    objects: List[ModelObject] = []
    sql: ModelSql



from pathlib import Path

try:
    from pyspark.sql import SparkSession
    from pyspark.sql.types import StructType, StructField, DataType, LongType, DoubleType, TimestampType, StringType, \
        BooleanType
except ImportError:
    raise Exception('Для запуска скрипта модели необходимо установить библиотеку pyspark')

from aw_client.models.model_schema import ModelObject


def get_virtual_object_dataframe(spark: SparkSession, model_object: ModelObject):
    """ """

    if not model_object.fields:
        raise Exception(f'Поля для виртуального объекта {model_object.model_name} не заданы')

    schema_fields = []
    for field in model_object.fields:
        schema_fields.append(
            StructField(
                name=field.model_name,
                dataType=get_spark_type_for_simple_type(field.simple_type),
                nullable=True
            )
        )

    return spark.createDataFrame(data=[], schema=StructType(schema_fields))


def get_spark_type_for_simple_type(simple_type: str) -> DataType:
    """ """
    if simple_type == 'number':
        return LongType()
    elif simple_type == 'float':
        return DoubleType()
    elif simple_type == 'date':
        return TimestampType()
    elif simple_type == 'bool':
        return BooleanType()
    else:
        return StringType()



from typing import List, Dict, Union, Optional, Any


class NamedObjectsBundle:
    """ """
    def __init__(self, objs: Dict[str, Any]):
        self._obj_list: List[Any] = [df for _, df in objs.items()]
        self._obj_named: Dict[str, Any] = objs

    def first(self) -> Optional[Any]:
        """ """
        return self._obj_list[0] if self._obj_list else None
    
    def as_list(self) -> List[Any]:
        return self._obj_list
    
    def as_named(self) -> Dict[str, Any]:
        return self._obj_named
    
    def __getitem__(self, item):
        if isinstance(item, int):
            # если item указано целым числом, то возвращаем по индексу
            return self._obj_list[item]
        
        if not item in self._obj_named:
            raise Exception(f'Данные объекта с именем "{item}" не найдена')
        return self._obj_named[item]
    
    def __iter__(self):
        return iter(self._obj_named.values())

    def __bool__(self):
        return len(self._obj_list) > 0
    
    def __len__(self):
        return len(self._obj_list)
    
    def items(self):
        return self._obj_named.items()



from typing import Optional

import abc
from pathlib import Path

import pandas
import httpx

from contextlib import contextmanager

from aw_client.models.model_schema import ModelSchema
from aw_client.domain import APIConfig


class DataMasterApi(metaclass=abc.ABCMeta):
    """
    Базовый класс для методов API дата мастера
    """
    class Error(Exception):
        """ """

    class MisconfiguredError(Error):
        """ Ошибка """

    def __init__(self, api_config: APIConfig):
        self.api_config = api_config

    @abc.abstractmethod
    def load_model(self, model_id: int, **options) -> pandas.DataFrame:
        """ """

    @abc.abstractmethod
    def model_schema(self, model_id: int) -> ModelSchema:
        """ """

    @abc.abstractmethod
    def load_model_object_data(self, folder: Path, model_id: int, model_object_name: str, limit: Optional[int]):
        """ """

    @contextmanager
    def get_http_client(self) -> httpx.Client:
        """ """
        if not self.api_config.aw_url:
            raise DataMasterApi.MisconfiguredError('Не указан URL к Analytic Workspace')

        if not self.api_config.token:
            raise DataMasterApi.MisconfiguredError('Не указано токен доступа к Analytic Workspace')

        headers = {"Authorization": f"Bearer {self.api_config.token}"}

        client = httpx.Client(base_url=self.api_config.aw_url, headers=headers)

        try:
            yield client
        finally:
            client.close()



from typing import Protocol, List, Any


class ModelObjectField(Protocol):
    """ """
    @property
    def simple_type(self) -> str:
        """ """


class ModelObject(Protocol):
    """ """
    @property
    def fields(self) -> List[ModelObjectField]:
        """ """


class Vault(Protocol):
    """ """
    def get(self, name: str) -> Any:
        """ """




from setuptools import setup, find_packages
import pathlib


here = pathlib.Path(__file__).parent.resolve()


long_description = (here / "README.md").read_text(encoding="utf-8")


setup(
    name='analytic_workspace_client',
    version='1.29.0rc1',

    description='Библиотека для подключения к Analytic Workspace',
    long_description=long_description,
    long_description_content_type="text/markdown",


    author='Analytic Workspace',
    author_email='aw_help@analyticworkspace.ru',
    url='https://analyticworkspace.ru/',

    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],

    python_requires='>=3.8,<4',

    package_dir={'': 'src'},
    packages=find_packages(where='src'),

    install_requires=[
        'python-dotenv>=1.0,<1.1',
        'httpx>=0.25,<1.0',
        'pandas',
        'pydantic>=1.10,<2.0',
        'colorama>=0.4,<0.5'
    ],

    extras_require={
        'dev': ['pyspark==3.5.0', 'pytest>=8.2,<8.3'],
        'ml': ['mlflow>=2.14,<2.15']
    },
    
    setup_requires=['wheel'],  
)



from typing import List, Tuple, Optional

import builtins


class CompiledModule:
    """ """
    def __init__(self, compiled_globals: dict):
        self.compiled_globals = compiled_globals

    def __contains__(self, item):
        return item in self.compiled_globals

    def __getattr__(self, item):
        return self.compiled_globals[item]

    def __getitem__(self, item):
        return self.compiled_globals[item]

    @property
    def is_empty(self) -> bool:
        """ Возвращает True, если в модуле нет кода (только одни комментарии, например) """
        return len(set(self.compiled_globals) - {'__builtins__'}) == 0


class ScriptCompiler:
    """ """

    class CompilerException(Exception):
        """ """

    # Список исключений
    class Misconfigured(CompilerException):
        """ """

    class ForbiddenImport(CompilerException):
        """ Исключение, которое возникает при наличии некорректных импортов в коде """

    class CannotCompile(CompilerException):
        """ Исключение, которое вызывается при невозможности компиляции кода """

    # Перечисление режимов работы компилятора
    MODE_ETL = 1
    MODE_VIRTUAL_OBJECT_SCHEMA = 2
    MODE_ETL_BLOCK = 3

    def __init__(self, mode: Optional[int] = None):
        self.mode = mode

    def compile(self, source_code: str, mode: int = None) -> CompiledModule:
        """ """
        try:
            byte_code = compile(source_code, filename='<string>', mode='exec')
        except SyntaxError as e:
            raise ScriptCompiler.CannotCompile(f'Ошибка компиляции: {e.msg} (line {e.lineno}): {e.text}')

        try:
            compiled_globals = self._get_globals(mode=mode if mode is not None else self.mode)
            exec(byte_code, compiled_globals)
        except Exception as e:
            raise ScriptCompiler.CompilerException(f'Ошибка компиляции: {e}')

        return CompiledModule(compiled_globals=compiled_globals)

    def _get_globals(self, mode: int) -> dict:
        """ """
        if mode == ScriptCompiler.MODE_ETL:
            safe_names, safe_modules = self._safe_globals_for_model_etl_script()
        elif mode == ScriptCompiler.MODE_ETL_BLOCK:
            safe_names, safe_modules = self._safe_globals_for_etl_block()
        elif mode == ScriptCompiler.MODE_VIRTUAL_OBJECT_SCHEMA:
            safe_names, safe_modules = self._safe_globals_for_virtual_object_schema_script()
        else:
            raise ScriptCompiler.Misconfigured(f'Указан некорректный режим компиляции {self.mode}')

        def safe_import(name, globals=None, locals=None, fromlist=(), level=0):
            """ """
            modules = name.split('.')
            if modules[0] not in safe_modules:
                raise ScriptCompiler.ForbiddenImport(f'Импорт модуля {modules[0]} запрещен')

            return __import__(name, globals, locals, fromlist, level)

        safe_builtins = {}
        for name in safe_names:
            safe_builtins[name] = getattr(builtins, name)

        safe_builtins['__import__'] = safe_import

        return {'__builtins__': safe_builtins}

    @staticmethod
    def _safe_globals_for_virtual_object_schema_script() -> Tuple[List[str], List[str]]:
        """ """
        safe_names = [
            'None', 'False', 'True', 'bool', 'bytes', 'chr', 'complex', 'float',
            'hex', 'id', 'int', 'str', 'getattr', 'setattr', 'delattr',
        ]

        safe_modules = ['pyspark', 'aw_etl']

        return safe_names, safe_modules

    @staticmethod
    def _safe_globals_full() -> Tuple[List[str], List[str]]:
        """ Возвращает полный набор разрешенных модулей """
        safe_names = [
            'None', 'False', 'True', 
            'abs', 'all', 'any', 'ascii', 
            'bin', 'bool', 'bytes', 'bytearray', 
            'callable', 'chr', 'classmethod', 'complex', 
            'delattr', 'dict', 'divmod', 
            'enumerate',
            'float', 'filter', 'format', 'frozenset',
            'getattr',
            'hasattr', 'hash', 'hex', 
            'id', 'int', 'isinstance', 'issubclass', 'iter',
            'len', 'list',
            'map', 'max', 'min',
            'next', 
            'object', 'oct', 'ord',
            'pow', 'print', 'property', 
            'range', 'repr', 'reversed', 'round', 
            'set', 'setattr', 'slice', 'sorted', 'staticmethod', 'str', 'sum', 'super',
            'type', 'tuple', 
            'vars',
            'zip', 
            '__build_class__', '__name__',
            

            'ArithmeticError', 'AssertionError', 'AttributeError', 'BaseException', 'BufferError', 'BytesWarning',
            'DeprecationWarning', 'EOFError', 'EnvironmentError', 'Exception', 'FloatingPointError', 'FutureWarning',
            'GeneratorExit', 'IOError', 'ImportError', 'ImportWarning', 'IndentationError', 'IndexError', 'KeyError',
            'KeyboardInterrupt', 'LookupError', 'MemoryError', 'NameError', 'NotImplementedError', 'OSError',
            'OverflowError', 'PendingDeprecationWarning', 'ReferenceError', 'RuntimeError', 'RuntimeWarning',
            'StopIteration', 'SyntaxError', 'SyntaxWarning', 'SystemError', 'SystemExit', 'TabError', 'TypeError',
            'UnboundLocalError', 'UnicodeDecodeError', 'UnicodeEncodeError', 'UnicodeError', 'UnicodeTranslateError',
            'UnicodeWarning', 'UserWarning', 'ValueError', 'Warning', 'ZeroDivisionError',
        ]

        safe_modules = [
            'pyspark', 'requests', 'pandas', 'numpy', 'aw_etl', 'pyparsing', 'pydantic',

            'mlflow', 'prophet', 'statmodels', 'torch', 'sklearn', 'numpy', 'catboost', 

            'array', 'calendar', 'codecs', 'collections', 'copy', 'csv', 'dataclasses', 'datetime', '_strptime',
            'decimal', 'enum', 'functools', 'hashlib', 'itertools', 'json', 'math', 'queue', 'random', 're',
            'statistics', 'string', 'time', 'urllib', 'xml',
            'zoneinfo', 'typing', 'uuid', 'logging',
        ]

        return safe_names, safe_modules

    @staticmethod
    def _safe_globals_for_model_etl_script() -> Tuple[List[str], List[str]]:
        """ """
        return ScriptCompiler._safe_globals_full()

    @staticmethod
    def  _safe_globals_for_etl_block() -> Tuple[List[str], List[str]]:
        """ """
        safe_names, safe_modules = ScriptCompiler._safe_globals_full()
        safe_modules.extend(['sqlglot', 'inspect'])

        return safe_names, safe_modules



from typing import Protocol, Optional

try:
    from pyspark.sql import SparkSession
except ImportError:
    raise Exception('Для использования Spark установите библиотеку с опцией [dev]: `pip install analytic-workspace-client[dev]`')


from aw_etl.models import Vault
from aw_etl.compiler import CompiledModule


class InvalidEtlBlock(Exception):
    """ """


class ETLBlockApplication(Protocol):
    """ """
    @property
    def spark(self) -> SparkSession:
        """ """

    @property
    def is_spark_initialized(self) -> bool:
        """ """

    @property
    def model_module(self) -> Optional[CompiledModule]:
        """ """

    @property
    def vault(self) -> Vault:
        """ """

    @property
    def run_mode(self) -> str:
        """ """

    


from typing import Optional
from pathlib import Path
import pandas
import io

from aw_client.data_master.base import DataMasterApi
from aw_client.exceptions import WrongDataMasterApiVersion
from aw_client.models.model_schema import ModelSchema


class DataMasterV0(DataMasterApi):
    """ """

    def load_model(self, model_id: int, **options) -> pandas.DataFrame:
        """ """
        with self.get_http_client() as client:
            url = f"data-master/v0/model/data?model_id={model_id}&format=csv"
            r = client.get(url, timeout=None)
            if not r.is_success:
                raise Exception(f"Ошибка загрузки данных модели: {r.text}")

        return pandas.read_csv(io.StringIO(r.content.decode("utf-8")))

    def model_schema(self, model_id: int) -> ModelSchema:
        raise WrongDataMasterApiVersion(f'Получение схемы модели доступно для AW с версии 1.19')

    def load_model_object_data(self, folder: Path, model_id: int, model_object_name: str, limit: Optional[int]):
        """ """
        raise WrongDataMasterApiVersion(f'Получение данных объекта модели доступно для AW с версии 1.19')



from typing import Optional, Any, List, Union

import datetime
from enum import Enum

from pydantic import BaseModel


class ETLBlockParamType(str, Enum):
    """ """
    STRING = 'string'
    TEXT = 'text'
    PASSWORD = 'password'
    SQL_TEXT = 'sql_text'
    NUMBER = 'number'
    FLOAT = 'float'
    BOOL = 'bool'
    DATE = 'date'
    DATETIME = 'datetime'
    SELECT = 'select'
    ACTION = 'action'


class ETLBlockParamGroupType(str, Enum):
    """ """
    GROUP = 'group'


class ETLBlockParamActionType(str, Enum):
    """ """
    ACTION = 'action'


# ----------------------------------------------------------------------------------------------------------------------
# Метаданные ETL блока
# ----------------------------------------------------------------------------------------------------------------------
class ETLBlockParam(BaseModel):
    """ Параметр ETL блока """
    code: str
    name: str
    type: ETLBlockParamType
    description: Optional[str] = None

    required: bool
    mult: bool
    domain: Optional[Any] = None
    extra: Optional[Any] = None


class ETLBlockParamGroup(BaseModel):
    """ Группа параметров ETL блока """
    code: str
    name: str

    type: ETLBlockParamGroupType
    description: Optional[str] = None
    view_options: Optional[dict] = None
    mult: bool
    params: List[Union['ETLBlockParam', 'ETLBlockParamAction', 'ETLBlockParamAction']] = []
    extra: Optional[Any] = None


class ETLBlockParamAction(BaseModel):
    code: str
    name: str
    type: ETLBlockParamActionType
    description: Optional[str] = None
    action: str
    extra: Optional[Any] = None


class ETLBlockMeta(BaseModel):
    """ Метаданные ETL-блока """
    uid: str  # уникальный идентификатор блока
    name: str  # название блока
    version: str  # версия блока
    description: str  # описание блока
    author: str  # автор блока
    updated_at: datetime.datetime  # дата и время последнего обновления
    params: List[Union[ETLBlockParam, ETLBlockParamAction, ETLBlockParamGroup]]
    engine_requires: Optional[List[str]] = []

    @property
    def verbose_name(self):
        return f'{self.name} v{self.version}'



import sys
import os
import datetime
from concurrent.futures import ThreadPoolExecutor

from importlib import import_module
from pathlib import Path
from argparse import ArgumentParser

import httpx
from colorama import Fore

from aw_client.models.model_schema import ModelObject, ModelSchema
from aw_client.session import Session

from aw_client.model_dev.cache import ModelDevCache
from aw_client.model_dev.application import Application
from aw_client.model_dev.virtual_objects import get_virtual_object_dataframe


try:
    import pyspark
    from pyspark.sql import SparkSession
    from pyspark.sql.types import StringType, LongType, DateType, TimestampType
    import pyspark.sql.functions as F
except ImportError:
    raise Exception('Для запуска скрипта модели необходимо установить библиотеку pyspark')


# ----------------------------------------------------------------------------------------------------------------------
# UDF функции
# ----------------------------------------------------------------------------------------------------------------------
def udf_aw_mask(text, mask_symb, pos):
    """ """
    return None if text is None else (mask_symb*pos + text[pos:])[:len(text)]


def udf_aw_cast_interval(dt_start: datetime.datetime, dt_finish: datetime.datetime, measure="sec"):
    """
    Возвращает интервал между двумя временными значениями (dt_finish - dt_start)
    в единицах, заданных параметром measure: [days|hours|minutes|sec|ms].
    В случае, если хотя бы одна из границ периода передана как None, возвращается None.
    В случае, если measure имеет недопустимое значение, возвращается None.
    """

    if dt_start is None or dt_finish is None:
        return None

    if not isinstance(dt_start, datetime.datetime):
        dt_start = datetime.datetime.combine(dt_start, datetime.time.min)

    if not isinstance(dt_finish, datetime.datetime):
        dt_finish = datetime.datetime.combine(dt_finish, datetime.time.min)

    delta = dt_finish - dt_start

    try:
        if measure == "days":
            return delta.days
        elif measure == "hours":
            return delta.days * 24 + delta.seconds // 3600
        elif measure == "minutes":
            return delta.days * 24 * 60 + delta.seconds // 60
        elif measure == "sec":
            return delta.days * 24 * 3600 + delta.seconds
        elif measure == "ms":
            return (delta.days * 24 * 3600 + delta.seconds) * 1000 + delta.microseconds // 1000
        else:
            return None
    except OverflowError:
        return None


if __name__ == '__main__':

    parser = ArgumentParser()
    parser.add_argument('file', type=str)
    parser.add_argument('--aw-url', dest='aw_url', type=str, required=True)
    parser.add_argument('--aw-token', dest='aw_token', type=str, required=True)
    parser.add_argument('--no-cache', dest='no_cache', action='store_true', default=False, required=False)
    parser.add_argument('--full', dest='full', action='store_true',  default=False, required=False)
    parser.add_argument('--cores', dest='cores', type=int, default=None)

    args, _ = parser.parse_known_args()

    script_path = Path(args.file)

    try:
        model_id = int(script_path.stem[6:])
    except Exception as e:
        print(Fore.RED + f'Не получить номер модели из названия файла {script_path}. '
                         f'Имя файла должно быть вида model_NNN.py, где вместо NNN ожидается идентификатор '
                         f'модели (например, model_123.py)')
        sys.exit(1)

    # Импортируем указанный файл
    print('--------------------------------------------------------------')
    print(f'Запускается ETL-скрипт с параметрами:')
    print(f'  - model_id: {model_id}')
    print(f'  - file: {script_path}')
    print(f'  - AW url: {args.aw_url}')
    print(f'  - Кеш данных: {"нет" if args.no_cache else "да"}')
    print(f'  - Полный набор данных из источника: {"да" if args.full else "нет"}')
    print(f'  - Использовать ядра: {"все" if args.cores is None else args.cores}')
    print('--------------------------------------------------------------\n')
    sys.path.insert(0, Path(args.file).parent.as_posix())
    etl_script_module = import_module(Path(args.file).stem)

    # Сессия для работы с AW
    session = Session(aw_url=args.aw_url, token=args.aw_token)

    try:
        if session.get_aw_version() < 2:
            print(Fore.RED + 'Версия AW должна быть как минимум 1.19' + Fore.RESET)
            sys.exit(-1)
    except httpx.HTTPError as e:
        print(Fore.YELLOW + f'Не удалось проверить версию AW на актуальность: {e}' + Fore.RESET)

    # Работа
    cache = ModelDevCache(
        cache_root_folder=Path(args.file).parent / '.aw-cache', aw_url=args.aw_url, model_id=model_id,
        is_full=args.full)

    # Сессия спарка
    master_url = 'local'
    if args.cores is not None:
        master_url = f'local[{args.cores}]'
    else:
        master_url = 'local[*]'
    spark = SparkSession.builder.config("spark.driver.host", "127.0.0.1").master(master_url).appName(f'model_{model_id}').getOrCreate()
    spark.sparkContext.setLogLevel('WARN')

    spark.udf.register('awmask', udf_aw_mask, StringType())
    spark.udf.register('awcastinterval', udf_aw_cast_interval, LongType())

    spark.conf.set('spark.sql.caseSensitive', False)
    spark.conf.set('spark.sql.files.maxPartitionBytes', 10000000)
    spark.conf.set('spark.sql.parquet.outputTimestampType', 'TIMESTAMP_MILLIS')
    spark.conf.set('spark.sql.parquet.datetimeRebaseModeInRead', 'CORRECTED')
    spark.conf.set('spark.sql.parquet.mergeSchema', 'true')
    spark.conf.set('spark.sql.parquet.datetimeRebaseModeInWrite', 'CORRECTED')

    # ------------------------------------------------------------------------------
    # Загрузка схемы модели
    # ------------------------------------------------------------------------------
    print('Загрузка схемы модели из AW... ', end=' ')
    try:
        model_schema = session.model_schema(model_id)
    except httpx.HTTPError as e:
        if cache.model_schema_path().exists() and not args.no_cache:
            print(Fore.YELLOW + f'Не удалось получить схему модели {model_id}: {e}. '
                                f'Для запуска скрипта используется схема модели из кеша.' + Fore.RESET)
            model_schema = ModelSchema.parse_file(cache.model_schema_path())
        else:
            print(Fore.RED + f'ОШИБКА: Не удалось получить схему модели {model_id}: {e}' + Fore.RESET)
            sys.exit(1)
    except Exception as e:
        print(Fore.RED + f'ОШИБКА: Не удалось получить схему модели {model_id}: {e}' + Fore.RESET)
        sys.exit(1)
    print(Fore.GREEN + 'OK' + Fore.RESET)

    cache_folder = cache.get_model_folder()
    os.makedirs(cache_folder, exist_ok=True)
    with open(cache.model_schema_path(), 'wt') as f:
        f.write(model_schema.json(ensure_ascii=False, indent=2))

    # ------------------------------------------------------------------------------
    # Загрузка данных объектов модели
    # ------------------------------------------------------------------------------
    def load_object_data(model_object: ModelObject, use_cache: bool):
        if cache.model_object_data_folder(model_object.model_name).exists() and use_cache and \
                cache.is_model_object_actual(model_object):
            print(Fore.RESET + f'Данные для {model_object.model_name} - ' + Fore.GREEN + 'OK (из кеша)' + Fore.RESET)
        else:
            print(Fore.RESET + f'Получение данных объекта {model_object.model_name} из AW ... ')
            try:
                session.load_model_object_data(folder=cache.model_source_folder(),
                                               model_id=model_id,
                                               model_object_name=model_object.model_name,
                                               limit=None if args.full else 10_000)
            except Exception as e:
                print(Fore.RED + f'Ошибка загрузки данных для {model_object.model_name}: {e}' + Fore.RESET)
                raise

            # сохраняем хеш для кешированных данных
            cache.set_model_object_hash(model_object)
            print(Fore.RESET + f'Данные для {model_object.model_name} - ' + Fore.GREEN + 'OK ' + Fore.RESET)
        return True

    with ThreadPoolExecutor(max_workers=4) as executor:
        non_virtual_objects = [o for o in model_schema.objects if o.type != 'virtual']
        for model_obj, obj_result in zip(non_virtual_objects, executor.map(
                lambda model_object: load_object_data(model_object, use_cache=not args.no_cache),
                non_virtual_objects
        )):
            if not obj_result:
                sys.exit(1)

    application = Application()
    # ------------------------------------------------------------------------------
    # Выполняем before_all
    # ------------------------------------------------------------------------------
    if hasattr(etl_script_module, 'before_all'):
        print(Fore.GREEN + 'Выполняется before_all(..)' + Fore.RESET)
        getattr(etl_script_module, 'before_all')(application, spark)

    # ------------------------------------------------------------------------------
    # Загрузка объектов модели
    # ------------------------------------------------------------------------------
    source_dfs = {}
    for model_object in model_schema.objects:
        if model_object.type == 'virtual':
            source_dfs[model_object.model_name] = get_virtual_object_dataframe(spark=spark, model_object=model_object)
        elif model_schema.sql.xsqls and model_object.model_name in model_schema.sql.xsqls:
            xsql = model_schema.sql.xsqls[model_object.model_name].replace('%WHERE%', '').replace(
                "%RUN_FOLDER%", cache.model_source_folder().as_posix())
            source_dfs[model_object.model_name] = spark.sql(xsql)
        else:
            try:
                source_dfs[model_object.model_name] = spark.read.parquet(
                    (cache.model_source_folder() / f'{model_object.model_name}.parquet').as_posix())
            except Exception as e:
                print(Fore.YELLOW + f'Не удалось загрузить данные для {model_object.model_name}: {e}' + Fore.RESET)
                # Молча проглотим ошибку, т.к. этот датафрейм может и не пригодиться в работе. Иначе,
                # возникнет ошибка ниже
                pass

    # ------------------------------------------------------------------------------
    # Выполняем функции из etl-скрипта
    # ------------------------------------------------------------------------------
    for model_object in model_schema.objects:
        method_name = f'after_load_{model_object.model_name}'
        if hasattr(etl_script_module, method_name) and model_object.model_name in source_dfs:
            print(Fore.GREEN + f'Выполняется after_load_{model_object.model_name}(..)' + Fore.RESET)
            source_dfs[model_object.model_name] = getattr(etl_script_module, method_name)(
                df=source_dfs[model_object.model_name],
                app=application,
                spark=spark
            )
            if not isinstance(source_dfs[model_object.model_name], pyspark.sql.DataFrame):
                raise Exception(f'Функция {method_name} должна вернуть объект типа pyspark.sql.DataFrame')

    # ------------------------------------------------------------------------------
    # Регистрация датафреймов из источников в spark-окружении
    # ------------------------------------------------------------------------------
    for obj_name, df in source_dfs.items():
        df.createOrReplaceTempView(obj_name)

    # ------------------------------------------------------------------------------
    # Выполняем итоговый SQL
    # ------------------------------------------------------------------------------
    print(Fore.GREEN + 'Выполняется итоговый SQL' + Fore.RESET)
    sql = model_schema.sql.sql.replace('parquet.`%RUN_FOLDER%/', '').replace('.parquet`', '').replace('%WHERE%', '')
    df_final = spark.sql(sql)

    # ------------------------------------------------------------------------------
    # Выполняем after_all
    # ------------------------------------------------------------------------------
    if hasattr(etl_script_module, 'after_all'):
        print(Fore.GREEN + 'Выполняется after_all(..)' + Fore.RESET)
        df_final = getattr(etl_script_module, 'after_all')(
            df=df_final,
            app=application,
            spark=spark
        )

    # ------------------------------------------------------------------------------
    # Конец
    # ------------------------------------------------------------------------------
    print(Fore.GREEN + 'ETL-скрипт успешно выполнен.' + Fore.RESET)



from typing import List, TypedDict, Optional

from dataclasses import dataclass, field


class ModelObjectSchemaField(TypedDict):
    """ """
    model_name: str
    simple_type: str


@dataclass
class ModelObjectTestData:
    """ """
    model_name: str
    rows: List[dict] = field(default_factory=list)
    schema: Optional[List[ModelObjectSchemaField]] = None
